"""
Provides the `InteractiveDeal` class which keeps an undo stack whenever its 
state is modified, making it easier to interact with the deal. 
"""

__all__ = ["InteractiveDeal"]

from endplay.interact.interactivedeal import InteractiveDeal