<!DOCTYPE html>
<html>
<head>
    <title>Dealer output</title>
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <link rel="preconnect" href="https://fonts.googleapis.com">
    <link rel="preconnect" href="https://fonts.gstatic.com" crossorigin>
    <link href="https://fonts.googleapis.com/css2?family=Overpass+Mono:wght@400&display=swap" rel="stylesheet">
    <style>
        html, body {
            padding: 0;
            margin: 0;
        }

        body {
            background-color: slategray;
        }

        .content {
            background-color: white;
            min-width: 600px;
            margin: 50px;
            padding: 20px;
            border: 5px solid #aaa;
            display: flex;
            flex-direction: column;
            align-items: center;
        }

        .deal {
            font-size: 16px;
            border: 1px solid #ddd;
            display: inline-flex;
            flex-flow: row wrap;
            width: 360px;
            justify-content: space-between;
            font-family: sans-serif;
            background-color: #fafafa;
            padding: 5px;
            border-radius: 5px;
        }

            .deal > div {
                padding: 10px;
                flex-basis: 100px;
                height: 100px;
            }

        .board-info {
            display: flex;
            flex-direction: column;
            align-items: center;
            justify-content: center;
            font-size: 20px;
            gap: 5px;
        }

        .hand {
            display: flex;
            flex-direction: column;
            justify-content: center;
            gap: 3px;
            font-family: "Overpass Mono", sans-serif;
            background-color: #f2efd0;
            overflow-wrap: anywhere;
            word-break: break-all;
            letter-spacing: 2px;
        }

        .hand-inline {
            display: flex;
            flex-direction: row;
            justify-content: center;
            gap: 3px;
            font-family: "Overpass Mono", sans-serif;
            background-color: #edfaa5;
            letter-spacing: 2px;
            padding: 5px;
        }

        .deal > .hand.north {
            border-radius: 5px 5px 0px 0px;
        }

        .deal > .hand.east {
            border-radius: 0px 5px 5px 0px;
        }

        .deal > .hand.south {
            border-radius: 0px 0px 5px 5px;
        }

        .deal > .hand.west {
            border-radius: 5px 0px 0px 5px;
        }

        .spades, .hearts, .diamonds, .clubs {
            text-indent: -16px;
            margin-left: 16px;
        }

            .spades:before, .hearts:before, .diamonds:before, .clubs:before {
                display: inline-block;
                width: 16px;
                text-indent: 0px;
            }

            .spades:before {
                content: "♠ ";
                color: #01039e;
            }

            .hearts:before {
                content: "♥ ";
                color: #a00203;
            }

            .diamonds:before {
                content: "♦ ";
                color: #c05e0a;
            }

            .clubs:before {
                content: "♣ ";
                color: #036a03;
            }

        .compass {
            display: flex;
            flex-flow: row wrap;
            color: white;
            background-color: #106610;
            align-items: center;
            font-weight: bold;
            user-select: none;
        }

            .compass .north {
                flex-basis: 100%;
                text-align: center;
            }

            .compass .west {
                flex-basis: 50%;
                text-align: left;
            }

            .compass .east {
                flex-basis: 50%;
                text-align: right;
            }

            .compass .south {
                flex-basis: 100%;
                text-align: center;
            }

        .dealer {
            text-decoration: underline;
        }

        .vul {
            color: #ff7777;
        }

        .nonvul {
            color: #77ff77;
        }

        .deal-compact {
            border-collapse: collapse;
        }

            .deal-compact .hand {
                padding: 10px;
            }

            .deal-compact th {
                color: white;
                font-weight: bold;
                background-color: #106610;
                font-family: sans-serif;
                padding: 5px;
            }

            .deal-compact td > div {
                width: 175px;
                height: 80px;
            }

            .deal-compact .value {
                background-color: #eee;
                height: 102px;
                display: flex;
                align-items: center;
                justify-content: center;
                font-family: sans-serif;
            }

        .deal-oneline {
            border-collapse: collapse;
        }

            .deal-oneline .hand {
                padding: 10px;
            }

            .deal-oneline th {
                color: white;
                font-weight: bold;
                background-color: #106610;
                font-family: sans-serif;
                padding: 5px;
            }

            .deal-oneline .value {
                background-color: #eee;
                height: 32px;
                display: flex;
                align-items: center;
                justify-content: center;
                font-family: sans-serif;
            }

        .valuebox {
            background-color: #eee;
            padding: 20px;
            border-radius: 10px;
            margin: 10px 50px;
            font-family: sans-serif;
        }
    </style>
</head>
<body>
    <div class="content">