"""
Endplay - A bridge tools library with generating, analysing and scoring.
Released under the MIT licence (see the LICENCE file provided with this distribution)
"""

import endplay._dds as _dds
from endplay.dds import *
from endplay.dealer import *
from endplay.interact import *
from endplay.parsers import *
from endplay.evaluate import *
from endplay.types import *
from endplay.config import \
	__version__, __version_info__, __author__, \
	__buildtime__, suppress_unicode