LIBRARY   dds
EXPORTS
   SetMaxThreads
   SetMaxThreads@4 = SetMaxThreads
   SetThreading
   SetThreading@4 = SetThreading
   SetResources
   SetResources@8 = SetResources
   GetDDSInfo
   GetDDSInfo@4 = GetDDSInfo
   FreeMemory
   FreeMemory@0 = FreeMemory
   ErrorMessage
   ErrorMessage@8 = ErrorMessage
   SolveBoard
   SolveBoard@116 = SolveBoard
   SolveBoardPBN
   SolveBoardPBN@132 = SolveBoardPBN
   CalcDDtable
   CalcDDtable@68 = CalcDDtable
   CalcDDtablePBN
   CalcDDtablePBN@84 = CalcDDtablePBN
   SolveAllBoards
   SolveAllBoards@8 = SolveAllBoards
   SolveAllBoardsBin
   SolveAllBoardsBin@8 = SolveAllBoardsBin
   SolveAllChunks
   SolveAllChunks@12 = SolveAllChunks
   SolveAllChunksBin
   SolveAllChunksBin@12 = SolveAllChunksBin
   SolveAllChunksPBN
   SolveAllChunksPBN@12 = SolveAllChunksPBN
   CalcAllTables
   CalcAllTables@20 = CalcAllTables
   CalcAllTablesPBN
   CalcAllTablesPBN@20 = CalcAllTablesPBN
   CalcPar
   CalcPar@76 = CalcPar
   SidesPar
   SidesPar@12 = SidesPar
   SidesParBin
   SidesParBin@12 = SidesParBin
   CalcParPBN
   CalcParPBN@92 = CalcParPBN
   Par
   Par@12 = Par
   DealerPar
   DealerPar@16 = DealerPar
   DealerParBin
   DealerParBin@12 = DealerParBin
   ConvertToDealerTextFormat
   ConvertToDealerTextFormat@8 = ConvertToDealerTextFormat
   ConvertToSidesTextFormat
   ConvertToSidesTextFormat@8 = ConvertToSidesTextFormat
   AnalysePlayBin
   AnalysePlayBin@524 = AnalysePlayBin
   AnalysePlayPBN
   AnalysePlayPBN@230 = AnalysePlayPBN
   AnalyseAllPlaysBin
   AnalyseAllPlaysBin@16 = AnalyseAllPlaysBin
   AnalyseAllPlaysPBN
   AnalyseAllPlaysPBN16 = AnalyseAllPlaysPBN
   
   
