1 VERSIONINFO
 FILEVERSION 2,9,0,0
 PRODUCTVERSION 2,9,0,0
 FILEFLAGSMASK 0x3fL
 FILEFLAGS 0x0L
 FILEOS 0x40004L
 FILETYPE 0x2L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "CompanyName", "http://privat.bahnhof.se/wb758135/"
            VALUE "FileDescription", "dds MinGW"
            VALUE "FileVersion", "2.9.0.0"
            VALUE "InternalName", "dds.dll"
            VALUE "LegalCopyright", "Copyright © 2006-2014 Bo Haglund / 2014-18 Bo Haglund Soren Hein"
            VALUE "OriginalFilename", "dds.dll"
            VALUE "ProductName", "Bridge double dummy solver"
            VALUE "ProductVersion", "2.9.0.0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END
